import os

from functional_notations import F

from cmx.backends.markdown import CommonMark
from . import data

doc = CommonMark(root=os.getcwd(), prefix=".")
md = doc
# todo: implement this
csv = F @ doc.csv
# todo: implement this



import os
from copy import deepcopy

from ..utils import get_block, is_subclass, to_snake
from functional_notations import _F
from ml_logger import ML_Logger

from . import components
from ..utils import dedent
from ..with_hack import SkipContextManager


def video(frames=None, *, src, window, **kwargs):
    """save video at filename.

    :param data:
    :param filename:
    :param kwargs:
    :return:
    """
    file_path, *query_strs = src.split('?')
    if frames is not None:
        window.logger.save_video(frames, file_path)
    if file_path.endswith("gif"):
        return components.Image(src=src, **kwargs)
    else:
        return components.Video(src=src, **kwargs)


class Figure(components.Figure):
    def __init__(self, image=None, src=None, title=None, caption=None, *, window, **kwargs):
        if not isinstance(image, components.Component):
            image = window.image(image, src, window=window, **kwargs)
        super().__init__(image=image, src=src, title=title, caption=caption, window=window)


class Image(components.Image):
    def __init__(self, image=None, src=None, *, window, normalize=False, **kwargs):
        if src is not None:
            file_path, *query_strs = src.split('?')
            if image is not None:
                window.logger.save_image(image, file_path, normalize=normalize)
            super().__init__(src=src, **kwargs)
        else:
            super().__init__(image, **kwargs)


class Savefig(components.Figure):
    def __init__(self, key, caption=None, width=None, height=None, zoom=None, window=None, **kwargs):
        file_path, *_ = key.split('?')
        super().__init__(src=key, width=width, height=height, caption=caption, zoom=zoom, window=window, **kwargs)
        self.window.logger.savefig(file_path, **kwargs)


USER = os.environ.get('USER', None)
PWD = os.environ.get('PWD', None)


class CommonMark(components.Article):
    __filename = None
    counter = 0

    @property
    def hide(self):
        """This is an amazing shortcut that alllows you to skip execution
        of a section during writing!

        But this currently interference with pydev, the debugging utility
        used in pyCharm. I have not tested this on dgp yet, but I expect
        similar issues.
        """
        from contextlib import ExitStack
        return ExitStack()

    @property
    def skip(self):
        """This is an amazing shortcut that alllows you to skip execution
        of a section during writing!

        But this currently interference with pydev, the debugging utility
        used in pyCharm. I have not tested this on dgp yet, but I expect
        similar issues.
        """
        return SkipContextManager(True)

    def __init__(self, filename=None, overwrite=True, root=None, prefix=None, logger=None):
        """
        Called by the module __init__.py to create a global document object.

        :param filename: the filename for the output markdown file.
        :param overwrite: Whether to append to the existing document, or to clear and
            start afresh
        :param logger: allow passing in an existing logger to log to a remote server instead.
            this is similar to a figure object as in matplotlib.
        """
        super().__init__(window={to_snake(k): v for k, v in globals().items() if is_subclass(v, components.Component)})
        self.window['video'] = video
        self.window.logger = logger or ML_Logger(root=root, prefix=prefix)

        # if logger.root_dir:
        #     cprint(f"cmx root directory: {logger.root_dir}", color="green")

        self.config(filename=filename, overwrite=overwrite)

    def config(self, filename=None, overwrite=True, src_prefix=None, logger=None):
        self.overwrite = overwrite
        self.window.logger = logger or self.window.logger
        # todo: for gist, the prefix needs to go into the `src` attribute.
        # self.window.src_prefix = src_prefix
        if filename:
            self.__filename = filename

            if self.overwrite:
                self.window.logger.log_text("", filename=self.filename, overwrite=True)

            from termcolor import cprint
            from urllib import parse
            if self.window.logger.root.startswith("http"):
                # just print the path, README.md should just show.
                cprint("File output at " + self.window.logger.get_dash_url() + " " + self.__filename, "green")
            else:
                cprint("File output at file://" + parse.quote(os.path.realpath(self.__filename)), "green")
        return self

    def new(self, filename=None, **kwargs):
        if filename:
            import os
            filename = os.path.abspath(filename)

        doc = deepcopy(self).config(filename=filename, **kwargs)
        return doc

    @property
    def filename(self):
        if self.__filename is None:
            import inspect
            filename = "cmx/"
            frame = inspect.currentframe()
            while "cmx/" in filename or "importlib" in filename or "contextlib" in filename:
                frame = frame.f_back
                filename, line_number, function_name, lines, index = inspect.getframeinfo(frame)

            # todo: '/__init__.py' instead?
            if filename.endswith('__init__.py'):
                self.__filename = filename[:-11] + "README.md"
            else:
                self.__filename = filename.replace(".py", ".md")

            # # on first write:
            if self.overwrite:
                self.clear()

            from termcolor import cprint
            from urllib import parse
            cprint("File output at file://" + parse.quote(self.__filename), "green")

        return self.__filename

    def write(self, text, overwrite=None):
        filename = self.filename
        if filename.startswith("/"):
            filename = "/" + filename
        self.window.logger.log_text(text, filename=filename, overwrite=overwrite)

    def clear(self):
        self.write("", overwrite=True)

    def __call__(self, *snippets, dedent=True, **kwargs):
        """output text"""
        t = components.Text(*snippets, dedent=dedent, **kwargs)
        self.children.append(t)
        return self

    def __matmul__(self, string_or_array):
        if isinstance(string_or_array, tuple):
            string, *rest = string_or_array
            return self(string, *rest)

        return self(string_or_array)

    md = __call__

    # def text(self, *args, sep=" ", end="\n"):
    #     t = components.Text(*args, sep=sep, end=end)
    #     self.children.append(t)
    #
    # def link(self, href=None):
    #     l = components.Link(href=href)
    #     self.children.append(l)

    # @property
    # def pre(self, ):
    #     def _pre(*args, sep=" ", lang=None):
    #         p = components.Pre(sep.join([str(a) for a in args]), lang=lang)
    #         self.children.append(p)
    #
    #     return _F(_pre, name="Pre")

    @property
    def yaml(self, data=None, **kwargs):
        def _yaml(data, **kwargs):
            import yaml
            s = yaml.dump(data).rstrip()
            return self.pre(s, lang="yaml")

        return _F(_yaml, name="yaml")

    @property
    def csv(self, ):
        def _csv(csv, show_index=False, **kwargs):
            return self.children.append(components.Table(csv, show_index=show_index, **kwargs))

        return _F(_csv, name="csv")

    def print(self, *args, sep=" ", end="\n"):
        if self.children and isinstance(self.children[-1], components.Print):
            self.children[-1].text += sep.join([str(a) for a in args]) + end
        else:
            self.children.append(components.Print(*args, sep=sep, end=end))
        print(*args, sep=sep, end=end)

    def flush(self, *args):
        self.write(self._md)
        self.children.clear()

    def __enter__(self):
        import inspect
        assert self.filename, "make sure that file is already set."

        prior_frame = inspect.currentframe().f_back
        filename, line_number, function_name, lines, index = inspect.getframeinfo(prior_frame)
        try:
            lines_in_block = get_block(filename, line_number + 1)
            text = "".join(lines_in_block)
            # todo: change to self.code(, lang="python", ...)
            self.children.append(components.Pre(dedent(text).rstrip(), lang="python"))
        except FileNotFoundError:
            print('in iPython session')
        return self

    __exit__ = flush


class Github(CommonMark):
    """uses tables for the layout"""
    pass


class Gist(CommonMark):
    """saves everything inside a folder"""
    pass



from cmx.backends.components import Article


def test_figure_row():
    doc = Article()
    table = doc.table()
    row = table.figure_row()
    row.figure(src="some_file.png", title="some title", caption="some text")
    row.figure(src="some_file.png", title="some title", caption="some text")
    row.figure(src="some_file.png", title="some title", caption="some text")
    row.figure(src="some_file.png", title="some title", caption="some text")

    row = table.figure_row()
    row.figure(src="some_file.png", title="some title", caption="some text")
    row.figure(src="some_file.png", title="some title", caption="some text")
    row.figure(src="some_file.png", title="some title", caption="some text")
    row.figure(src="some_file.png", title="some title", caption="some text")

    print(table._md)
    assert table._md == """
| **some title** | **some title** | **some title** | **some title** |
|:--------------:|:--------------:|:--------------:|:--------------:|
| ![some_file.png](some_file.png) | ![some_file.png](some_file.png) | ![some_file.png](some_file.png) | ![some_file.png](some_file.png) |
| some text | some text | some text | some text |
| **some title** | **some title** | **some title** | **some title** |
| ![some_file.png](some_file.png) | ![some_file.png](some_file.png) | ![some_file.png](some_file.png) | ![some_file.png](some_file.png) |
| some text | some text | some text | some text |
"""[1:]



from cmx import doc


def test_str():
    a = """
    here is a string
    """
    b = """
here is a string
"""
    doc @ a
    assert doc._md == b
    doc.children.clear()


def test_print():
    for i in range(2):
        doc.print(i, "<")

    target = """
```
0 <
1 <
```
"""[1:]
    print(doc._md)
    assert doc._md == target
    doc.children.clear()


def test_table():
    table = doc.table()

    with table.figure_row() as row:
        row.figure(src="some_file.png", title="some title", caption="some text")
        row.figure(src="some_file.png", title="some title", caption="some text")
        row.figure(src="some_file.png", title="some title", caption="some text")
        row.figure(src="some_file.png", title="some title", caption="some text")

    with table.figure_row() as row:
        row.figure(src="some_file.png", title="some title", caption="some text")
        row.figure(src="some_file.png", title="some title", caption="some text")
        row.figure(src="some_file.png", title="some title", caption="some text")
        row.figure(src="some_file.png", title="some title", caption="some text")

    target = """
| **some title** | **some title** | **some title** | **some title** |
|:--------------:|:--------------:|:--------------:|:--------------:|
| ![some_file.png](some_file.png) | ![some_file.png](some_file.png) | ![some_file.png](some_file.png) | ![some_file.png](some_file.png) |
| some text | some text | some text | some text |
| **some title** | **some title** | **some title** | **some title** |
| ![some_file.png](some_file.png) | ![some_file.png](some_file.png) | ![some_file.png](some_file.png) | ![some_file.png](some_file.png) |
| some text | some text | some text | some text |
"""[1:]

    print(doc._md)

    assert doc._md == target
    doc.children.clear()


def test_image():
    from skimage import data

    img = data.astronaut()
    doc.image(img)
    print(doc._md)
    doc.flush()


def test_image_src():
    from skimage import data

    img = data.camera()
    doc.image(img, f"figures/reach.png?ts={doc.now()}")
    print(doc._md)
    doc.flush()


def test_figure_row():
    doc @ """
    ## Test Figure Row
    """
    with doc:
        from skimage import data

        img = data.coins()

    with doc, doc.table() as table:
        with table.figure_row() as row:
            row.figure(img, src=f"figures/reach.png?ts={doc.now()}", title="Before Init", caption="this is the details")
            row.figure(img, src=f"figures/reach.png?ts={doc.now()}", title="Before Init", caption="this is the details")
            row.figure(img, src=f"figures/reach.png?ts={doc.now()}", title="Before Init", caption="this is the details")
            row.figure(img, src=f"figures/reach.png?ts={doc.now()}", title="Before Init", caption="this is the details")



def is_empty(line):
    return len(line.strip()) == 0


def get_indent(text):
    for line in text.split("\n"):
        if line:
            return len(line) - len(line.lstrip())
    return 0


def dedent(text):
    lines = text.split('\n')
    for l in lines:
        if l:
            break
    indent = get_indent(l)
    return '\n'.join([l[indent:] for l in lines])


def get_block(filename, line_number):
    import linecache
    current_line_number = 0
    lines, indent = [], None
    with open(filename, 'r') as f:
        while True:
            line = f.readline()
            if line is "":  # this is the end of lines
                break
            current_line_number += 1
            if current_line_number < line_number:
                continue
            current_indent = get_indent(line)
            if line == "\n":
                lines.append(line)
            else:
                if indent is None:
                    indent = current_indent
                elif current_indent < indent:
                    break
                    # print(lines, sep="\n")
                    # print(current_indent, indent)
                lines.append(line)

    return lines


def is_subclass(obj, cls):
    try:
        return issubclass(obj, cls)
    except TypeError:
        return False


import re


def to_snake(name):
    name = re.sub('(.)([A-Z][a-z]+)', r'\1_\2', name)
    return re.sub('([a-z0-9])([A-Z])', r'\1_\2', name).lower()



import sys


class SkipException(Exception):
    pass


class SkipContextManager:
    def __init__(self, skip):
        """
        :param skip: bool, skip code execution if True.
        """
        self.skip = skip

    def __enter__(self):
        if self.skip:
            sys.settrace(lambda *args, **keys: None)
            frame = sys._getframe(1)
            frame.f_trace = self.trace

    def trace(self, frame, event, arg):
        raise SkipException()

    def __exit__(self, type, value, traceback):
        if type is None:
            return  # No exception
        if issubclass(type, SkipException):
            return True  # Suppress special SkipWithBlock exception


if __name__ == '__main__':
    with SkipContextManager(skip=True) \
            as this_works:
        print('In the with block')  # Won't be called
    print('Out of the with block')



from functional_notations import _F


@_F
def csv(csv_str, **kwargs):
    """
    :param csv_str:
    :param kwargs:
    :return: DataFrame object
    """
    data = None
    # under construction
    return data


@_F
def yaml(yaml_str, _ruamel, **kwargs):
    """

    :param yaml_str:
    :param _ruamel:
    :param kwargs:
    :return: DataFrame object
    """
    data = None
    return data


@_F
def table(yaml_str, _ruamel, **kwargs):
    """

    :param yaml_str:
    :param _ruamel:
    :param kwargs:
    :return: DataFrame object
    """
    data = None
    return data



from os import path
from setuptools import setup, find_packages

with open(path.join(path.abspath(path.dirname(__file__)), 'VERSION'), encoding='utf-8') as f:
    version = f.read()

with open(path.join(path.abspath(path.dirname(__file__)), 'README'), encoding='utf-8') as f:
    long_description = f.read()

setup(
    name='cmx',
    packages=find_packages(),
    install_requires=[
        "sanic==20.9.0",
        "killport",
    ],
    description='long_description',
    long_description=long_description,
    author='Ge Yang<ge.ike.yang@gmail.com>',
    url='https://github.com/cmx/cmx-python',
    author_email='ge.ike.yang@gmail.com',
    package_data={'cmx': ['cmx/*.*', 'cmx/**/*.*', 'tassa', 'tassa/*.*']},
    version=version,
)



from types import SimpleNamespace
from functional_notations import F

from waterbear import Bear

from .. import utils
import pandas as pd
from io import StringIO

from cmx.utils import is_subclass, to_snake


def attrs(class_name=None, **kwargs):
    attrs_str = f'class="{class_name}"' if class_name else ""
    attrs_str += " ".join([k.replace('_', "-") + f'="{str(v)}"' for k, v in kwargs.items()])

    return attrs_str


def styles(**kwargs):
    return " ".join([k.replace('_', "-") + f':{str(v)};' for k, v in kwargs.items()])


class FuncCall(SimpleNamespace):
    name = None
    args = tuple()
    kwargs = {}


class Component(object):
    tag = None
    data = None
    class_name = None
    window = Bear()

    def now(self, fmt=None):
        from datetime import datetime
        now = datetime.now().astimezone()
        return now.strftime(fmt) if fmt else now

    def __init__(self, tag=None, children=None, window=None, **kwargs):
        self.tag = tag or self.tag
        self.kwargs = kwargs
        self.style = {}
        self.children = children or []
        if window:
            _window = self.window.copy()
            _window.update(window)
            self.window = _window

    def _callback(self, item, *args, window=None, **kwargs):
        # change to self.comp_list attribute to make explicit.
        if window:
            _window = self.window.copy()
            _window.update(window)
            window = _window
        else:
            window = self.window
        Comp = window[item]
        component = Comp(*args, window=window, **kwargs)
        self.children.append(component)
        return component

    def __getattribute__(self, item):
        try:
            return object.__getattribute__(self, item)
        except AttributeError as e:
            def proxy_fn(*args, **kwargs):
                """this leaks memory :)"""
                return self._callback(item, *args, **kwargs)

            return proxy_fn
            # return F @ proxy_fn

    @property
    def _attrs(self):
        return attrs(class_name=self.class_name, **self.kwargs)

    @property
    def _md(self):
        return self._html + "\n"

    @property
    def _html(self):
        # todo: add styles to this.
        return f"<{self.tag} {self._attrs}>{''.join([b._html for b in self.children if b is not None])}</{self.tag}>"

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        pass


class Container(Component):
    """does not support _md and _html"""

    @property
    def _md(self):
        raise RuntimeError(f'Container{self.__name__} does not support markdown generation directly')

    @property
    def _html(self):
        raise RuntimeError(f'Container{self.__name__} does not support html generation directly')


class Div(Component):
    tag = "div"


class Span(Component):
    tag = "span"

    def __init__(self, *args, sep=" ", dedent=None, **kwargs):
        super().__init__(**kwargs)
        self.text = sep.join([str(a) for a in args])
        if dedent:
            self.text = utils.dedent(self.text)

    @property
    def _md(self):
        return self.text

    @property
    def _html(self):
        return f"<{self.tag}>{self.text}</{self.tag}>"


Text = Span


class Bold(Span):
    tag = "b"

    @property
    def _md(self):
        return f"**{super()._md}**"


class Pre(Component):
    tag = "pre"

    def __init__(self, text, lang=None, **kwargs):
        super().__init__(**kwargs)
        self.text = text
        self.lang = lang

    @property
    def _md(self):
        text = self.text + ('' if self.text.endswith('\n') else '\n')
        if not text.startswith('\n'):
            text = '\n' + text
        return f"```{self.lang if self.lang else ''}{text}```\n"

    @property
    def _html(self):
        # todo: support language strings
        if self.lang:
            segs = [
                '<pre>',
                f'<code class="{self.lang}">',
                f'{self.text}',
                f'</code>',
                '</pre>'
            ]
        else:
            segs = [
                '<pre>',
                f'{self.text}',
                '</pre>'
            ]
        return "\n".join(segs) + "\n"


class Link(Component):
    tag = "span"

    def __init__(self,href="", text="", **kwargs):
        super().__init__(**kwargs)
        self.text = text
        self.href = href

    @property
    def _md(self):
        return f'[{self.text}]({self.href})'

    @property
    def _html(self):
        return f'<a href="{self.href}">{self.text}</a>'


class Img(Component):
    tag = "img"
    zoom = None

    def __init__(self, src=None, zoom=None, alt=None, **kwargs):
        super().__init__(**kwargs)
        self.src = src
        self.alt = alt
        if zoom is not None:
            self.style = {"zoom": zoom}
        self.zoom = zoom

    @property
    def _md(self):
        if self.zoom is None:
            return f"![{self.src}]({self.alt or self.src})"
        else:
            return self._html

    @property
    def _html(self):
        # prevent stretched when inside flex-box.
        return f'<img style="{styles(align_self="center", **self.style)}" src="{self.src}" {self._attrs}/>'


class Image(Img):
    """Advanced Image with Data handling"""
    data = None

    def __init__(self, image=None, src=None, normalize=False, **kwargs):
        if image is None:
            super().__init__(src=src, **kwargs)
        else:
            import numpy as np
            # todo: need to support 16 and 32 bit images, especially depth images.
            self.data = np.array(image).astype(np.uint8)
            super().__init__(src=self.base64, **kwargs)

    @property
    def base64(self):
        # if self.data is not None:
        assert self.data is not None
        from io import BytesIO
        from PIL import Image as pImage
        import base64

        with BytesIO() as buf:
            pImage.fromarray(self.data).save(buf, "png")
            return "data:image/png;base64," + base64.b64encode(buf.getvalue()).decode('utf-8')


class Video(Component):
    def __init__(self, src=None, width=320, height=240, controls=True, **kwargs):
        super().__init__(**kwargs)
        self.src = src
        self.width = width
        self.height = height
        self.controls = controls

    @property
    def _html(self):
        return utils.dedent(f"""
        <video width="{self.width}" height="{self.height}" controls="{str(self.controls).lower()}">
          <source src="{self.src}" type="video/mp4">
          Your browser does not support the video tag.
        </video>
        """)


class Figure(Component):
    tag = "div"

    def __init__(self, image=None, src=None, title=None, caption=None, **kwargs):
        super().__init__(**kwargs)
        self.src = src
        if isinstance(image, Component):
            self.img = image
        else:
            self.img = self.window.img(image=image, src=src, styles=dict(margin="0.5em"), **kwargs)
        self.title = title
        self.caption = caption

        self.children = [self.title, self.img, self.caption]

    @property
    def _html(self):
        if self.title and self.caption:
            return f"""<table><tr><td rowspan="2">{self.img._html}</td><td>{self.title}</td></tr><tr><td>{self.caption}</td></tr></table>"""
        elif self.caption:
            return f"""<table><tr><td rowspan="2">{self.img._html}</td><td>{self.caption}</td></tr></table>"""
        elif self.title:
            return f"""<table><tr><th>{self.title}</th></tr><tr><td>{self.img._html}</td></tr></table>"""
        return self.img._html


class Row(Component):
    styles = dict(display="flex",
                  flex_direction="row",
                  item_align="center", )

    def __init__(self, wrap=False, styles={}, **kwargs):
        super().__init__(**kwargs)
        if wrap is not None:
            wrap = "wrap" if wrap else "nowrap"

        self.styles = dict(flex_wrap=wrap, **Row.styles)
        self.styles.update(styles)

    @property
    def _html(self):
        # use children's HTML instead of markdown.
        return f'<div style="{styles(**self.styles)}">{"".join([c._html for c in self.children])}</div>'


# todo: use table component for images
# fixme: Not Implemented
# class TableCell(Component):
#     pass
# class TableRow(Component):
#     pass
# class TableColumn(Component):
#     pass

class FigureRow(Container):
    def __init__(self, header=None, n_cols=None, **kwargs):
        super().__init__(**kwargs)
        # self.header = header
        # todo: add fixed n_cols support
        self.n_cols = n_cols
        self.titles = []
        self.images = []
        self.footers = []

    def figure(self, image=None, src=None, title=None, caption=None, **kwargs):
        # todo: if image is Image/Video/Component
        self.titles.append(title if title is None else Bold(title))
        self.footers.append(caption if caption is None else Span(caption))
        self.image(image=image, src=src, **kwargs)

    def video(self, frames=None, src=None, title=None, caption=None, **kwargs):
        self.titles.append(title if title is None else Bold(title))
        self.footers.append(caption if caption is None else Span(caption))
        v = self.window.video(frames=frames, src=src, window=self.window, **kwargs)
        self.children.append(v)

    def column(self, title=None, text=None, footer=None):
        self.titles.append(title if title is None else Bold(title))
        self.footers.append(footer if footer is None else Span(footer))
        self.children.append(Text(text))

    def savefig(self, key, title=None, caption=None, **kwargs):
        self.titles.append(title if title is None else Bold(title))
        self.footers.append(caption if caption is None else Span(caption))
        self.children.append(self.window.savefig(key, window=self.window, **kwargs))

    @property
    def rows(self):
        non_empty_rows = []
        for r in [self.titles, self.children, self.footers]:
            for item in r:
                if item is not None:
                    non_empty_rows.append(r)
                    break
        return non_empty_rows


# todo: HashTable: use (row, col, row_span, col_span) to record cells.
class Table(Component):
    data = None

    def __init__(self, table=None, show_index=None, format="github", sep=",*", **kwargs):
        super().__init__(**kwargs)
        self.show_index = show_index
        self.kwargs = kwargs
        self.format = format
        if table is None:
            pass
        elif isinstance(table, str):
            self.data = pd.read_csv(StringIO(table), sep=sep)
        elif isinstance(table, pd.DataFrame):
            self.data = table
        else:
            self.data = pd.DataFrame(table)

    @property
    def _md(self):
        # todo: pad columns (done automatically?)
        # we organize by rows. Columns has to be contained in a row.
        if not self.children:
            return self.data.to_markdown(index=self.show_index,
                                         tablefmt=self.format, **self.kwargs) + "\n"
        rows = []
        for child in self.children:
            if isinstance(child, FigureRow):
                rows.extend(child.rows)
            else:
                rows.append(child.children)
        _md_str = ""
        for i, r in enumerate(rows):
            _md_str += "| " + " | ".join([' ' if c is None else c._md.strip() for c in r]) + " |\n"
            if i == 0:
                _md_str += "|:" + ":|:".join(['-' if c is None else '-' * len(c._md.strip()) for c in r]) + ":|\n"
        return _md_str

    @property
    def _html(self):
        return self.data.to_html(index=self.show_index, **self.kwargs)


# todo: should be Body, b/c Html includes body and head
class Html(Component):
    tag = "body"
    window = Bear(**{to_snake(k): v for k, v in globals().items() if is_subclass(v, Component)})


class Article(Html):
    tag = "article"
    class_name = "commonmark"
    window = Bear(**{to_snake(k): v for k, v in globals().items() if is_subclass(v, Component)})

    @property
    def _md(self):
        return "\n".join([c._md for c in self.children])


class Print(Pre):
    def __init__(self, *args, sep=" ", end="\n"):
        super().__init__(sep.join([str(a) for a in args]) + end)


from sanic import Sanic, Request, Websocket

app = Sanic("CMX Backend")


@app.websocket("/socket-echo")
async def echo(request: Request, ws: Websocket):
    async for msg in ws:
        await ws.send(msg)


@app.route("/log")
async def log(request: Request, ws: Websocket):
    pass